import os
import sys
import logging
import logging.handlers


def setup_logging(*, 
                 logfile=None, logdir="logs", prog_name="gway",
                 loglevel="INFO", pattern=None, backup_count=7):
    """Globally configure logging with optional rotating file and console handlers."""
    
    # Determine log level
    loglevel = getattr(logging, str(loglevel).upper(), logging.INFO)
    
    # Ensure logs directory exists if file logging is enabled
    if logfile:
        os.makedirs(logdir, exist_ok=True)
        if not os.path.isabs(logfile):
            logfile = os.path.join(os.getcwd(), logdir, logfile)
    
    # Define default pattern (now using funcName instead of filename). Only time, no date.
    # TODO: Replace funcName with the name of the python script followed by : then the line number
    pattern = pattern or '%(asctime)s %(levelname)s [%(name)s] %(filename)s:%(lineno)d - %(message)s'

    # Clear existing handlers to avoid duplicates
    logger = logging.getLogger()
    if logger.handlers:
        for handler in logger.handlers[:]:
            logger.removeHandler(handler)
    logger.setLevel(loglevel)

    # File handler (rotating daily at midnight)
    if logfile:
        file_handler = logging.handlers.TimedRotatingFileHandler(
            logfile, when='midnight', interval=1, backupCount=backup_count, encoding='utf-8'
        )
        file_handler.setLevel(loglevel)
        file_handler.setFormatter(logging.Formatter(pattern, datefmt='%H:%M:%S'))
        logger.addHandler(file_handler)

    # Initial log message
    sep = "-" * 70
    cmd_args = " ".join(sys.argv[1:])
    logger.info(f"\n{sep}\n> {prog_name or '%prog'} {cmd_args}\n{sep}")
    logger.info(f"Loglevel set to {loglevel} ({logging.getLevelName(loglevel)})")

    return logger



# TODO: Prioritize and analye code and its TODOs. 
# Indicate the most urgent/important TODO to implement and propose a standard implementation for it.
# Fully write all modified functions preserving the coding style. 

import os
import sys
import time
import inspect
import logging
import asyncio
import threading
import importlib
import functools

from .sigils import Resolver
from .structs import Results


class Gateway(Resolver):
    _builtin_cache = None
    _thread_local = threading.local()

    # TODO: Allow Gateway to receive a "buffer" parameter which may be a filename
    # in a string or Path object, or a file object or stream. 

    # TODO: Rewrite IO functions on the Gateway instance to use this buffer by default.

    def __init__(self, verbose=False, **kwargs):

        # TODO: Support a "project" param that will import all functions
        # in that project as additional builtin functions (first-level attributes)

        self._cache = {}
        self._async_threads = []
        self.base_path = os.path.dirname(os.path.dirname(__file__))
        self.logger = logging.getLogger("gw")
        if not verbose:
            self.verbose =  lambda *_, **__: None
        elif verbose is True:
            self.verbose =  lambda *args, **kwargs: self.info(*args, **kwargs)

        # Thread-local context/results
        if not hasattr(Gateway._thread_local, "context"):
            Gateway._thread_local.context = {}
        if not hasattr(Gateway._thread_local, "results"):
            Gateway._thread_local.results = Results()

        # Inject initial context if provided
        Gateway._thread_local.context.update(kwargs)

        # Aliases for convenience
        self.context = Gateway._thread_local.context
        self.results = Gateway._thread_local.results

        # Resolver setup
        super().__init__([
            ('results', self.results),
            ('context', self.context),
            ('env', os.environ),
        ])

        # Cache builtins once
        if Gateway._builtin_cache is None:
            builtins_module = importlib.import_module("gway.builtins")
            Gateway._builtin_cache = {
                name: obj for name, obj in inspect.getmembers(builtins_module)
                if inspect.isfunction(obj)
                and not name.startswith("_")
                and inspect.getmodule(obj) == builtins_module
            }

        self._builtin_functions = Gateway._builtin_cache.copy()

    def _wrap_callable(self, func_name, func_obj):
        @functools.wraps(func_obj)
        def wrapped(*args, **kwargs):
            try:
                self.debug(f"Call <{func_name}>: {args=} {kwargs=}")
                sig = inspect.signature(func_obj)
                bound_args = sig.bind_partial(*args, **kwargs)
                bound_args.apply_defaults()
                # self.debug(f"Context before argument injection: {self.context}")

                for param in sig.parameters.values():
                    if param.name not in bound_args.arguments and param.kind not in (param.VAR_POSITIONAL, param.VAR_KEYWORD):
                        default_value = param.default
                        if isinstance(default_value, str) and default_value.startswith("[") and default_value.endswith("]"):
                            resolved = self.resolve(default_value)
                            bound_args.arguments[param.name] = resolved

                # Use explicit kwargs provided to override existing context
                for key, value in bound_args.arguments.items():
                    if isinstance(value, str):
                        resolved = self.resolve(value)
                        bound_args.arguments[key] = resolved
                    self.context[key] = bound_args.arguments[key]

                # Prepare args and kwargs for function call
                args_to_pass = []
                kwargs_to_pass = {}
                for param in sig.parameters.values():
                    self.debug(f"Preparing {param=}")
                    if param.kind == param.VAR_POSITIONAL:
                        bound_val = bound_args.arguments.get(param.name, ())
                        self.debug(f"Kind == VAR_POSITIONAL {bound_val=} -> extend args")
                        args_to_pass.extend(bound_val)
                    elif param.kind == param.VAR_KEYWORD:
                        bound_val = bound_args.arguments.get(param.name, {})
                        self.debug(f"Kind == VAR_KEYWORD {bound_val=} -> extend kwargs")
                        kwargs_to_pass.update(bound_val)
                    elif param.name in bound_args.arguments:
                        bound_val = bound_args.arguments[param.name]
                        if param.default == bound_val:
                            found_val = self.find_value(param.name)
                            self.debug(f"Checking override: {bound_val=} == {param.default=} => {found_val=}")
                            if found_val is not None and found_val != bound_val:
                                self.info(f"Injected {param.name}={found_val} overrides default {bound_val=}")
                                bound_val = found_val
                        else:
                            self.debug(f"Value for {param.name} differs from default "
                                       f"({param.default=}); using provided {bound_val=}")
                        kwargs_to_pass[param.name] = bound_val
                    else:
                        self.debug(f"No preparation procedure matched for {param.name}")

                # Handle coroutine function
                if inspect.iscoroutinefunction(func_obj):
                    thread = threading.Thread(
                        target=self._run_coroutine,
                        args=(func_name, func_obj, args_to_pass, kwargs_to_pass),
                        daemon=True
                    )
                    self._async_threads.append(thread)
                    thread.start()
                    return f"[async task started for {func_name}]"

                # Call synchronous function
                result = func_obj(*args_to_pass, **kwargs_to_pass)

                # Handle coroutine result from sync function
                if inspect.iscoroutine(result):
                    thread = threading.Thread(
                        target=self._run_coroutine,
                        args=(func_name, result),
                        daemon=True
                    )
                    self._async_threads.append(thread)
                    thread.start()
                    return f"[async coroutine started for {func_name}]"

                # Store result
                short_key = func_name.split("_")[-1] if "_" in func_name else func_name
                long_key = func_name.split("_", 1)[-1] if "_" in func_name else func_name
                self.info(f"Stored {short_key=} {long_key=} {result=}")
                self.results.insert(short_key, result)
                if long_key != short_key:
                    self.results.insert(long_key, result)
                if isinstance(result, dict):
                    self.context.update(result)

                return result

            except Exception as e:
                self.error(f"Error in '{func_name}': {e}")
                raise

        return wrapped
        
    def _run_coroutine(self, func_name, coro_or_func, args=None, kwargs=None):
        try:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            # Determine if it's already a coroutine object
            if asyncio.iscoroutine(coro_or_func):
                result = loop.run_until_complete(coro_or_func)
            else:
                result = loop.run_until_complete(coro_or_func(*(args or ()), **(kwargs or {})))

            self.results.insert(func_name, result)
            if isinstance(result, dict):
                self.context.update(result)
        except Exception as e:
            self.error(f"Async error in {func_name}: {e}")
        finally:
            loop.close()

    # TODO: Log as self.critical when coroutine has been scheduled but the program terminates
    # without until being called. If possible, have a way for us to have a callback
    # so that we have the option to call until ourselves (under limited conditions.)

    def until(self, lock_file=None, lock_url=None, lock_pypi=False):
        from .watchers import watch_file, watch_url, watch_pypi_package
        def shutdown(reason):
            self.warning(f"{reason} triggered async shutdown.")
            os._exit(1)

        watchers = [
            (lock_file, watch_file, "Lock file"),
            (lock_url, watch_url, "Lock url"),
            (lock_pypi if lock_pypi is not False else None, watch_pypi_package, "PyPI package")
        ]
        for target, watcher, reason in watchers:
            if target:
                self.info(f"Setup watcher for {reason}")
                if target is True and lock_pypi:
                    target = "gway"
                watcher(target, on_change=lambda r=reason: shutdown(r))
        try:
            while any(thread.is_alive() for thread in self._async_threads):
                time.sleep(0.1)
        except KeyboardInterrupt:
            self.critical("KeyboardInterrupt received. Exiting immediately.")
            os._exit(1)

    def __getattr__(self, name):
        # Delegate standard logger methods to self.logger
        if hasattr(self.logger, name) and callable(getattr(self.logger, name)):
            return getattr(self.logger, name)

        # Builtin function?
        if name in self._builtin_functions:
            func = self._wrap_callable(name, self._builtin_functions[name])
            setattr(self, name, func)
            return func

        # Cached project?
        if name in self._cache: return self._cache[name]

        # Attempt dynamic project loading via load_project
        try:
            project_obj = self.load_project(project_name=name)
            return project_obj
        except Exception as e:
            raise AttributeError(f"Project or builtin '{name}' not found: {e}")

    def load_project(self, project_name: str, root: str = "projects"):
        # Replace hyphens with underscores in module names
        project_name = project_name.replace("-", "_")
        project_path = gw.resource(root, *project_name.split("."))
        self.debug(f"Load {project_name=} under {root=} -> {project_path}")
        load_mode = None

        if os.path.isdir(project_path) and os.path.isfile(os.path.join(project_path, "__init__.py")):
            # It's a package
            project_file = os.path.join(project_path, "__init__.py")
            module_name = project_name.replace(".", "_")
            load_mode = "package"
        else:
            # It's a single module
            project_file = str(project_path) + ".py"
            if not os.path.isfile(project_file):
                raise FileNotFoundError(f"Project file or package not found: {project_file}")
            module_name = project_name.replace(".", "_")
            load_mode = "module"

        spec = importlib.util.spec_from_file_location(module_name, project_file)
        if spec is None or spec.loader is None:
            raise ImportError(f"Could not load spec for {project_name}")

        project_module = importlib.util.module_from_spec(spec)
        sys.modules[module_name] = project_module  # Important for relative imports
        spec.loader.exec_module(project_module)

        if load_mode == "module":
            project_functions = {
                name: obj for name, obj in inspect.getmembers(project_module, inspect.isfunction)
                if not name.startswith("_") and obj.__module__ == project_module.__name__
            }
        else:
            project_functions = {
                name: obj for name, obj in inspect.getmembers(project_module, inspect.isfunction)
                if not name.startswith("_")
            }

        # Wrap project functions in a project object
        project_obj = type(project_name, (), {})()
        for func_name, func_obj in project_functions.items():
            wrapped_func = self._wrap_callable(f"{project_name}.{func_name}", func_obj)
            setattr(project_obj, func_name, wrapped_func)

        # Cache and return
        self._cache[project_name] = project_obj
        return project_obj
    

# This line allows using "from gway import gw" everywhere else
gw = Gateway()



import os
import sys
import json
import time
import inspect
import argparse
from typing import get_origin, get_args, Literal, Union

from .logging import setup_logging
from .builtins import abort
from .environs import load_env, get_base_client, get_base_server
from .gateway import gw, Gateway


def cli_main():
    """Main CLI entry point."""
    parser = argparse.ArgumentParser(description="Dynamic Project CLI")
    parser.add_argument("-a", dest="all", action="store_true", help="Return all results, not just the last one")
    parser.add_argument("-b", dest="batch", type=str, help="Run commands from a batch script")
    parser.add_argument("-c", dest="client", type=str, help="Specify client environment")
    parser.add_argument("-d", dest="debug", action="store_true", help="Enable debug logging")
    parser.add_argument("-e", dest="expression", type=str, help="Return resolved sigil at the end")
    parser.add_argument("-j", dest="json", action="store_true", help="Output result(s) as JSON")
    parser.add_argument("-s", dest="server", type=str, help="Specify server environment")
    parser.add_argument("-t", dest="timed", action="store_true", help="Enable timing")
    parser.add_argument("-v", dest="verbose", action="store_true", help="Verbose mode")
    parser.add_argument("-x", dest="callback", type=str, help="Execute a callback per command")
    parser.add_argument("commands", nargs=argparse.REMAINDER, help="Project/Function command(s)")
    args = parser.parse_args()

    # TODO: Can we make -v also work if we pass a string to it?
    # Then we can pass it like this: Gateway(verbose=args.verbose)

    loglevel = "DEBUG" if args.debug else "INFO"
    setup_logging(logfile="gway.log", loglevel=loglevel)
    start_time = time.time() if args.timed else None

    env_root = os.path.join(gw.base_path, "envs")
    client_name = args.client or get_base_client()
    server_name = args.server or get_base_server()

    load_env("client", client_name, env_root)
    load_env("server", server_name, env_root)

    if args.batch:
        command_sources, comments = load_batch(args.batch)
        gw.info(f"Comments in batch script:\n{chr(10).join(comments)}")
    else:
        if not args.commands:
            parser.print_help()
            sys.exit(1)
        command_sources = chunk_command(args.commands)

    if args.callback:
        callback = gw[args.callback]
        all_results, last_result = process_commands(command_sources, callback=callback)
    else:
        all_results, last_result = process_commands(command_sources)

    # Print all results immediately if --all is set
    if args.all:
        for result in all_results:
            if args.json:
                print(json.dumps(result, indent=2, default=str))
            elif result is not None:
                gw.info(f"Result:\n{result}")
                print(result)

    # Final result resolution
    output = Gateway(**last_result).resolve(args.expression) if args.expression else last_result

    # Only print final result if --all wasn't used
    if not args.all:
        if args.json:
            print(json.dumps(output, indent=2, default=str))
        elif output is not None:
            gw.info(f"Last function result:\n{output}")
            print(output)
        else:
            gw.info("No results returned.")

    if start_time:
        print(f"\nElapsed: {time.time() - start_time:.4f} seconds")


def process_commands(command_sources, callback=None, **context):
    """Shared logic for executing CLI or batch commands with optional per-node callback."""
    from gway import gw, Gateway

    all_results = []
    last_result = None

    gw = Gateway(**context) if context else gw

    for chunk in command_sources:
        if not chunk:
            continue

        gw.debug(f"Processing chunk: {chunk}")

        # Invoke callback if provided
        if callback:
            callback_result = callback(chunk)
            if callback_result is False:
                gw.debug(f"Skipping chunk due to callback: {chunk}")
                continue
            elif isinstance(callback_result, list):
                gw.debug(f"Callback replaced chunk: {callback_result}")
                chunk = callback_result
            elif callback_result is None or callback_result is True:
                pass  # continue with original chunk
            else:
                abort(f"Invalid callback return value for chunk: {callback_result}")

        if not chunk:
            continue

        raw_first_token = chunk[0]
        normalized_first_token = raw_first_token.replace("-", "_")
        remaining_tokens = chunk[1:]

        current_project_obj = None
        func_tokens = []
        project_functions = {}

        try:
            current_project_obj = getattr(gw, normalized_first_token)
            if callable(current_project_obj):
                project_functions = {raw_first_token: current_project_obj}
                func_tokens = [raw_first_token] + remaining_tokens
            else:
                project_functions = {
                    name: func for name, func in vars(current_project_obj).items()
                    if callable(func) and not name.startswith("_")
                }
                func_tokens = remaining_tokens or abort(f"No function specified for project '{raw_first_token}'")
        except AttributeError:
            try:
                builtin_func = getattr(gw.builtin, normalized_first_token)
                if callable(builtin_func):
                    project_functions = {raw_first_token: builtin_func}
                    func_tokens = [raw_first_token] + remaining_tokens
                else:
                    abort(f"Unknown command or project: {raw_first_token}")
            except AttributeError:
                if current_project_obj:
                    project_functions = {
                        name: func for name, func in vars(current_project_obj).items()
                        if callable(func) and not name.startswith("_")
                    }
                    func_tokens = [raw_first_token] + remaining_tokens
                else:
                    abort(f"Unknown project, builtin, or function: {raw_first_token}")

        raw_func_name = func_tokens[0]
        normalized_func_name = raw_func_name.replace("-", "_")
        func_args = func_tokens[1:]

        func_obj = project_functions.get(raw_func_name) or project_functions.get(normalized_func_name)
        if not func_obj:
            abort(f"Function '{raw_func_name}' not found.")

        func_parser = argparse.ArgumentParser(prog=raw_func_name)
        add_function_args(func_parser, func_obj)
        parsed_args = func_parser.parse_args(func_args)

        final_args, final_kwargs = prepare_arguments(parsed_args, func_obj)

        try:
            result = func_obj(*final_args, **final_kwargs)
            last_result = result
            all_results.append(result)
        except Exception as e:
            gw.exception(e)
            abort(f"Unhandled {type(e).__name__} in {func_obj.__name__}")

    return all_results, last_result

def prepare_arguments(parsed_args, func_obj):
    """Prepare *args and **kwargs for a function call."""
    func_args = []
    func_kwargs = {}
    extra_kwargs = {}

    for name, value in vars(parsed_args).items():
        param = inspect.signature(func_obj).parameters.get(name)
        if param is None:
            continue
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            func_args.extend(value or [])
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            if value:
                for item in value:
                    if '=' not in item:
                        abort(f"Invalid kwarg format '{item}'. Expected key=value.")
                    k, v = item.split("=", 1)
                    extra_kwargs[k] = v
        else:
            func_kwargs[name] = value

    return func_args, {**func_kwargs, **extra_kwargs}


def load_batch(batch_filename):
    """Load commands and comments from a .gws file."""
    commands = []
    comments = []

    if not os.path.isabs(batch_filename):
        candidate_names = [batch_filename]
        if not os.path.splitext(batch_filename)[1]:
            candidate_names += [f"{batch_filename}.gws", f"{batch_filename}.txt"]
        for name in candidate_names:
            batch_path = gw.resource("scripts", name)
            if os.path.isfile(batch_path):
                break
        else:
            abort(f"Batch script not found in scripts/: tried {candidate_names}")
    else:
        batch_path = batch_filename
        if not os.path.isfile(batch_path):
            abort(f"Batch file not found: {batch_path}")

    gw.info(f"Loading commands from batch: {batch_path}")

    with open(batch_path) as f:
        for line in f:
            stripped_line = line.strip()
            if stripped_line.startswith("#"):
                comments.append(stripped_line)
            elif stripped_line:
                commands.append(stripped_line.split())

    return commands, comments


def chunk_command(args_commands):
    """Split args.commands into logical chunks without breaking quoted arguments."""
    chunks = []
    current_chunk = []

    for token in args_commands:
        if token in ('-', ';'):
            if current_chunk:
                chunks.append(current_chunk)
                current_chunk = []
        else:
            current_chunk.append(token)

    if current_chunk:
        chunks.append(current_chunk)

    return chunks


def show_functions(functions: dict):
    """Display a formatted view of available functions."""
    print("Available functions:")
    for name, func in functions.items():
        args_list = []
        for param in inspect.signature(func).parameters.values():
            if param.default != inspect.Parameter.empty:
                args_list.append(f"--{param.name} {param.default}")
            else:
                args_list.append(f"--{param.name} <required>")
        args_preview = " ".join(args_list)

        doc = ""
        if func.__doc__:
            doc_lines = [line.strip() for line in func.__doc__.splitlines()]
            doc = next((line for line in doc_lines if line), "")

        print(f"  > {name} {args_preview}")
        if doc:
            print(f"      {doc}")


def add_function_args(subparser, func_obj):
    """Add the function's arguments to the CLI subparser."""
    sig = inspect.signature(func_obj)
    for arg_name, param in sig.parameters.items():
        if param.kind == inspect.Parameter.VAR_POSITIONAL:
            subparser.add_argument(arg_name, nargs='*', help=f"Variable positional arguments for {arg_name}")
        elif param.kind == inspect.Parameter.VAR_KEYWORD:
            subparser.add_argument('--kwargs', nargs='*', help='Additional keyword arguments as key=value pairs')
        else:
            arg_name_cli = f"--{arg_name.replace('_', '-')}"
            if param.annotation == bool or isinstance(param.default, bool):
                group = subparser.add_mutually_exclusive_group(required=False)
                group.add_argument(arg_name_cli, dest=arg_name, action="store_true", help=f"Enable {arg_name}")
                group.add_argument(f"--no-{arg_name.replace('_', '-')}", dest=arg_name, action="store_false", help=f"Disable {arg_name}")
                subparser.set_defaults(**{arg_name: param.default})
            else:
                arg_opts = get_arg_options(arg_name, param, gw)
                subparser.add_argument(arg_name_cli, **arg_opts)


def get_arg_options(arg_name, param, gw=None):
    """Infer argparse options from parameter signature."""
    opts = {}
    annotation = param.annotation
    default = param.default

    origin = get_origin(annotation)
    args = get_args(annotation)
    inferred_type = str

    if origin == Literal:
        opts["choices"] = args
        inferred_type = type(args[0]) if args else str
    elif origin == Union:
        non_none = [a for a in args if a is not type(None)]
        if len(non_none) == 1:
            inner_param = type("param", (), {"annotation": non_none[0], "default": default})
            return get_arg_options(arg_name, inner_param, gw)
        elif all(a in (str, int, float) for a in non_none):
            inferred_type = str
    elif annotation != inspect.Parameter.empty:
        inferred_type = annotation

    opts["type"] = inferred_type

    if default != inspect.Parameter.empty:
        if isinstance(default, str) and default.startswith("[") and default.endswith("]") and gw:
            try:
                default = gw.resolve(default)
            except Exception as e:
                print(f"Failed to resolve default for {arg_name}: {e}")
        opts["default"] = default
    else:
        opts["required"] = True

    return opts
